/**
****************************************************************************************************
****************************************************************************************************
@file: camera.cpp
@brief initialization and manipulating with camera
****************************************************************************************************
***************************************************************************************************/
#include "camera.h"

/**
****************************************************************************************************
@brief Initialize camera, sets all vectors to identity
***************************************************************************************************/
TCamera::TCamera()
{
  pos.LoadIdentity();
  rot.LoadIdentity();
}

/**
****************************************************************************************************
@brief Rotate camera around axis(relative)
@param angle rotation angle in degrees
@param axis rotation axis (can be A_X, A_Y, A_Z)
@return new transformation matrix
****************************************************************************************************/
TMatrix TCamera::Rotate(GLfloat angle, GLint axis)
{
  switch(axis)
  {
    case A_X: rot.x += angle;
        break;
    case A_Y: rot.y += angle;
        break;
    case A_Z: rot.z += angle;
        break;
    default:
        break;
  }
	UpdateMatrix();
	return transform;
}

/**
****************************************************************************************************
@brief Rotate camera around axis(absolute)
@param angle rotation angle in degrees
@param axis rotation axis (can be A_X, A_Y, A_Z)
@return new transformation matrix
****************************************************************************************************/
TMatrix TCamera::RotateAbs(GLfloat angle, GLint axis)
{
  switch(axis)
  {
    case A_X: rot.x = angle;
        break;
    case A_Y: rot.y = angle;
        break;
    case A_Z: rot.z = angle;
        break;
    default:
        break;
  }
	UpdateMatrix();
	return transform;
}

/**
****************************************************************************************************
@brief Move camera to new position(relative)
@param wx new X-coordinate
@param wy new Y-coordinate
@param wz new Z-coordinate
@return new transformation matrix
****************************************************************************************************/
TMatrix TCamera::Move(GLfloat wx, GLfloat wy, GLfloat wz)
{
  pos.x += wx;
  pos.y += wy;
  pos.z += wz;
	UpdateMatrix();
	return transform;
}

/**
****************************************************************************************************
@brief Move camera to new position(absolute)
@param wx new X-coordinate
@param wy new Y-coordinate
@param wz new Z-coordinate
@return new transformation matrix
****************************************************************************************************/
TMatrix TCamera::MoveAbs(GLfloat wx, GLfloat wy, GLfloat wz)
{
  pos.x = wx;
  pos.y = wy;
  pos.z = wz;
	UpdateMatrix();
	return transform;
}

/**
****************************************************************************************************
@brief Look camera at target.
@param wx target X-coordinate
@param wy target Y-coordinate
@param wz target Z-coordinate
@return new transformation matrix
****************************************************************************************************/
TMatrix TCamera::LookAt(GLfloat wx, GLfloat wy, GLfloat wz)
{
  look.x = wx;
  look.y = wy;
  look.z = wz;
	UpdateMatrix();
	return transform;
}


/**
****************************************************************************************************
@brief Update camera matrix by translating and rotating viewport
****************************************************************************************************/
void TCamera::UpdateMatrix()
{
	glLoadIdentity();
  //move camera absolute or
  if(look.IsEmpty())
  {
    glTranslatef(pos.x,pos.y,pos.z);
    glRotatef(rot.x,1.0f,0.0f,0.0f);
    glRotatef(rot.y,0.0f,1.0f,0.0f);
  }
  //by looking position
  else
  {
    gluLookAt(pos.x, pos.y, pos.z,
              look.x, look.y, look.z,
              0.0, 1.0, 0.0);
  }
	glGetFloatv(GL_MODELVIEW_MATRIX, transform);
}
