/**
****************************************************************************************************
****************************************************************************************************
@file light.cpp
@brief light class and light settings - implementation
****************************************************************************************************
***************************************************************************************************/
#include "light.h"

/**
****************************************************************************************************
@brief Make empty light: zero index, position at 0,0,0, without shadow
***************************************************************************************************/
TLight::TLight()
{
  light = 0;
  pos.LoadIdentity();
  shadow = false;
  fbo = 0;
}

/**
****************************************************************************************************
@brief Delete dynamic data
***************************************************************************************************/
TLight::~TLight()
{
	///delete framebuffer
	if(fbo != 0)
		glDeleteFramebuffers(1, &fbo);
}

/**
****************************************************************************************************
@brief Change light's component color
@param component light color component (can be AMBIENT,DIFFUSE,SPECULAR)
@param color color (as RGB TVector)
***************************************************************************************************/
void TLight::ChangeColor(GLint component, TVector color)
{
  switch(component)
  {
    case AMBIENT:
        glLightfv(light + GL_LIGHT0, GL_AMBIENT, color);
				ambient = color;
        break;
    case DIFFUSE:
        glLightfv(light + GL_LIGHT0, GL_DIFFUSE, color);
				diffuse = color;
        break;
    case SPECULAR:
        glLightfv(light + GL_LIGHT0, GL_SPECULAR, color);
				specular = color;
        break;
  };
}


/**
****************************************************************************************************
@brief Set all light parameters
@param lights light index in OpenGL
@param amb ambient color (RGB TVector)
@param diff diffuse color (RGB TVector)
@param spec specular color (RGB TVector)
@param lpos light position (XYZ TVector)
@param _radius light attenuation radius
***************************************************************************************************/
void TLight::Set(GLint lights, TVector amb, TVector diff, TVector spec, TVector lpos, float _radius)
{
  light = lights;

  //light settings
  ambient = amb;
  diffuse = diff;
  specular = spec;
  pos = lpos;
	radius = _radius;

  glLightfv(light + GL_LIGHT0, GL_AMBIENT, amb);
  glLightfv(light + GL_LIGHT0, GL_DIFFUSE, diff);
  glLightfv(light + GL_LIGHT0, GL_SPECULAR, spec);
  glLightfv(light + GL_LIGHT0, GL_POSITION,lpos);

	//light sphere
	q = gluNewQuadric();
  gluQuadricNormals(q, GLU_SMOOTH);

  shadow = false;
}



 /**
****************************************************************************************************
@brief  Draws current light position
@param transformMatrix current camera transformation matrix
***************************************************************************************************/
void TLight::DrawLight(TMatrix &transformMatrix)
{
	glPushMatrix();
	glLoadMatrixf(transformMatrix);
	glTranslatef(pos.x,pos.y,pos.z);
	glUseProgram(0);
	glBindTexture(GL_TEXTURE_2D,0);
	glColor3f(diffuse.x,diffuse.y,diffuse.z);
	gluSphere(q,0.5f,8,8);
  glPopMatrix();
}

/**
****************************************************************************************************
@brief Set light shadow
@param _shadow_size shadow map size (must be power of two)
@param _shadow_intensity mixing intensity (0 - transparent, 1 - opaque)
@param  _shadow shadows on/off
***************************************************************************************************/
void TLight::SetShadow(GLint _shadow_size, GLfloat _shadow_intensity, bool _shadow)
{
  //shadow properties
  shadow = _shadow;
  shadow_size = _shadow_size;
  shadow_intensity = _shadow_intensity;

  ///Use bias matrix to normalize texture coords for shadow map
  biasMatrix = TMatrix(0.5f, 0.0f, 0.0f, 0.0f,
                       0.0f, 0.5f, 0.0f, 0.0f,
                       0.0f, 0.0f, 0.5f, 0.0f,
                       0.5f, 0.5f, 0.5f, 1.0f);
}

/**
****************************************************************************************************
@brief Get shadow cache
@return unique shadow cache name for future use
***************************************************************************************************/
string TLight::GetShadowCache()
{
	stringstream shad_cache_name;
	shad_cache_name<<"shadow"<<light;
  return shad_cache_name.str();
}
