/**
****************************************************************************************************
****************************************************************************************************
@file: load3DS.cpp
@brief loads 3DS object from file
Uses lib3DS to load 3D objects/scenes: http://code.google.com/p/lib3ds/
****************************************************************************************************
***************************************************************************************************/
#include "object.h"

/*
//vertex attributes location
const int vertLoc = 0;
const int normLoc = 1;
const int coordLoc = 2;   //free attribute for nVidia cards (they don't allow reuse reserved index)
*/
/**
****************************************************************************************************
@brief Creates object directly from 3DS mesh
@param mesh 3DS mesh
@return pointer to vertex buffer with data
****************************************************************************************************/
VBO TObject::Create(Lib3dsMesh *mesh)
{  
  name = mesh->name;
  scale.Set(1.0,1.0,1.0);
  mat = NULL;
  shadow_cast = true;
  refl_size = 256;
  draw_object = true;
  type = EXTERN;
  drawmode = GL_TRIANGLES;
	element_indices = false;

	//count faces
	vbo.indices = mesh->nfaces;

  // Allocate memory for our vertices and normals
  TVertex *vertices = new TVertex[vbo.indices * 3];
  TVertex *normals = new TVertex[vbo.indices * 3];
	TCoord *texcoords = new TCoord[vbo.indices * 3];
 
  unsigned int FinishedFaces = 0;
	lib3ds_mesh_calculate_vertex_normals(mesh, &normals[FinishedFaces*3]);
  // Loop through every face
  for(unsigned curr_face = 0; curr_face < mesh->nfaces; curr_face++)
  {
    Lib3dsFace *face = &mesh->faces[curr_face];
    for(unsigned i = 0; i < 3; i++)
    {
			//vertices
			memcpy(&vertices[FinishedFaces*3 + i], mesh->vertices[face->index[i]], sizeof(TVertex) );
			//texture coordinates (if present)
			if(mesh->texcos != NULL)
				memcpy(&texcoords[FinishedFaces*3 + i], mesh->texcos[face->index[i]], sizeof(TCoord) );
    }
    FinishedFaces++;
  }
	//swap Y and Z coordinate in normals and vertices (3ds BUG)
	for(unsigned i=0; i<vbo.indices * 3; i++)
	{
		std::swap(*(vertices[i]+1),*(vertices[i]+2));
		*(vertices[i]+2) *= -1.0;
		std::swap(*(normals[i]+1),*(normals[i]+2));
		*(normals[i]+2) *= -1.0;
	}

  //Vertex array
  glGenVertexArrays(1, &vbo.vao);
  glBindVertexArray(vbo.vao);

  //Allocate and assign three VBO to our handle (vertices, normals and texture coordinates)
  glGenBuffers(4, vbo.buffer);

  //store vertices into buffer
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_VERTEX]);
  glBufferData(GL_ARRAY_BUFFER, sizeof(TVertex) * 3 * vbo.indices, vertices, GL_STATIC_DRAW);
#ifdef GL3
  // vertices are on index 0 and contains three floats per vertex
  glVertexAttribPointer(GLuint(0), 3, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(0);
#else
  glVertexPointer(3, GL_FLOAT, 0, 0);
  glEnableClientState(GL_VERTEX_ARRAY);
#endif
 
  //store normals into buffer
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_NORMAL]);
  glBufferData(GL_ARRAY_BUFFER, sizeof(TVertex) * 3 * vbo.indices, normals, GL_STATIC_DRAW);
#ifdef GL3
  // normals are on index 1 and contains three floats per vertex
  glVertexAttribPointer(GLuint(1), 3, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(1);
#else
  glNormalPointer(GL_FLOAT, 0, 0);
  glEnableClientState(GL_NORMAL_ARRAY);
#endif

  //store texture coordinates
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_TEXCOORD]);    
  glBufferData(GL_ARRAY_BUFFER, sizeof(TCoord) * 3 * vbo.indices, texcoords, GL_STATIC_DRAW);
#ifdef GL3
  //coordinates are on index 2 and contains two floats per vertex
  glVertexAttribPointer(GLuint(2), 2, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(2);
#else
  glTexCoordPointer(2, GL_FLOAT, 0, 0);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
#endif

  // Clean up our allocated memory
  delete [] vertices;
  delete [] normals;
	delete [] texcoords;

  //return VBO structure
  return vbo;
}



/**
****************************************************************************************************
@brief Creates object from external 3DS file
@param _name object name
@param _file file with object data in 3DS format
@param load has object data been loaded before?
@return success/fail of loading
****************************************************************************************************/
VBO TObject::Create(const char *_name, const char *_file, bool load)
{  
  name = _name;
  scale.Set(1.0,1.0,1.0);
  mat = NULL;
  shadow_cast = true;
  refl_size = 256;
  draw_object = true;
  type = EXTERN;
  drawmode = GL_TRIANGLES;
	element_indices = false;

  //don't load object if it has been loaded before
  if(!load) 
  {
    //set VAO and VBO from existing data
    glBindVertexArray(vbo.vao);
    glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_VERTEX]);
#ifdef GL3
    // vertices are on index 0 and contains three floats per vertex
    glVertexAttribPointer(GLuint(0), 3, GL_FLOAT, GL_FALSE, 0, 0);
    glEnableVertexAttribArray(0);
#else
    glVertexPointer(3, GL_FLOAT, 0, 0);
    glEnableClientState(GL_VERTEX_ARRAY);
#endif

    glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_NORMAL]);
#ifdef GL3
    // normals are on index 1 and contains three floats per vertex
    glVertexAttribPointer(GLuint(1), 3, GL_FLOAT, GL_FALSE, 0, 0);
    glEnableVertexAttribArray(1);
#else
    glNormalPointer(GL_FLOAT, 0, 0);
    glEnableClientState(GL_NORMAL_ARRAY);
#endif

    glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_TEXCOORD]); 
#ifdef GL3
    //coordinates are on index 2 and contains two floats per vertex
    glVertexAttribPointer(GLuint(2), 2, GL_FLOAT, GL_FALSE, 0, 0);
    glEnableVertexAttribArray(2);
#else
    glTexCoordPointer(2, GL_FLOAT, 0, 0);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
#endif

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo.buffer[P_INDEX]);
    return vbo;
  }

	/////////////////////////////////////////////////////////////////////////////
	//Load 3DS file using lib3ds
	cout<<"Loading object "<<_file<<"...";
	Lib3dsFile *model;
	model = lib3ds_file_open(_file);
	if(!model)
  {
    ShowMessage("Cannot open 3DS file!",false);
    return VBO();
  }

	//count faces
  vbo.indices = 0;
  // Loop through every mesh
	for(int i=0; i < model->nmeshes; i++)
  {
     // Add the number of faces this mesh has to the total faces
		 vbo.indices += model->meshes[i]->nfaces;
  }

  // Allocate memory for our vertices and normals
  TVertex *vertices = new TVertex[vbo.indices * 3];
  TVertex *normals = new TVertex[vbo.indices * 3];
	TCoord *texcoords = new TCoord[vbo.indices * 3];
 
  unsigned int FinishedFaces = 0;
  // Loop through all the meshes
  for(int curr_mesh = 0; curr_mesh < model->nmeshes; curr_mesh++)
  {
		Lib3dsMesh *mesh = model->meshes[curr_mesh];
		lib3ds_mesh_calculate_vertex_normals(mesh, &normals[FinishedFaces*3]);
    // Loop through every face
    for(unsigned curr_face = 0; curr_face < mesh->nfaces; curr_face++)
    {
      Lib3dsFace *face = &mesh->faces[curr_face];
      for(unsigned i = 0; i < 3; i++)
      {
				//vertices
				memcpy(&vertices[FinishedFaces*3 + i], mesh->vertices[face->index[i]], sizeof(TVertex) );
				//texture coordinates
				if(mesh->texcos != NULL)
					memcpy(&texcoords[FinishedFaces*3 + i], mesh->texcos[face->index[i]], sizeof(TCoord) );
      }
      FinishedFaces++;
    }
  }

	//swap Y and Z coordinate in normals and vertices (3ds BUG)
	for(unsigned i=0; i<vbo.indices * 3; i++)
	{
		std::swap(*(vertices[i]+1),*(vertices[i]+2));
		*(vertices[i]+2) *= -1.0;
		std::swap(*(normals[i]+1),*(normals[i]+2));
		*(normals[i]+2) *= -1.0;
	}

  //Vertex array
  glGenVertexArrays(1, &vbo.vao);
  glBindVertexArray(vbo.vao);

  //Allocate and assign three VBO to our handle (vertices, normals and texture coordinates)
  glGenBuffers(4, vbo.buffer);

  //store vertices into buffer
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_VERTEX]);
  glBufferData(GL_ARRAY_BUFFER, sizeof(TVertex) * 3 * vbo.indices, vertices, GL_STATIC_DRAW);
#ifdef GL3
  // vertices are on index 0 and contains three floats per vertex
  glVertexAttribPointer(GLuint(0), 3, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(0);
#else
  glVertexPointer(3, GL_FLOAT, 0, 0);
  glEnableClientState(GL_VERTEX_ARRAY);
#endif
 
  //store normals into buffer
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_NORMAL]);
  glBufferData(GL_ARRAY_BUFFER, sizeof(TVertex) * 3 * vbo.indices, normals, GL_STATIC_DRAW);
#ifdef GL3
  // normals are on index 1 and contains three floats per vertex
  glVertexAttribPointer(GLuint(1), 3, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(1);
#else
  glNormalPointer(GL_FLOAT, 0, 0);
  glEnableClientState(GL_NORMAL_ARRAY);
#endif

  //store texture coordinates
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_TEXCOORD]);    
  glBufferData(GL_ARRAY_BUFFER, sizeof(TCoord) * 3 * vbo.indices, texcoords, GL_STATIC_DRAW);
#ifdef GL3
  //coordinates are on index 2 and contains two floats per vertex
  glVertexAttribPointer(GLuint(2), 2, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(2);
#else
  glTexCoordPointer(2, GL_FLOAT, 0, 0);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
#endif

  // Clean up our allocated memory
  delete [] vertices;
  delete [] normals;
	delete [] texcoords;
 
  // We no longer need lib3ds
  lib3ds_file_free(model);
  model = NULL;

	cout<<"Done(vertices: "<<vbo.indices*3<<", faces: "<<vbo.indices<<")\n";

  //return VBO structure
  return vbo;
}
