/**
****************************************************************************************************
****************************************************************************************************
@file mat.cpp
@brief Simple mathematical library to work with vectors and matrices
****************************************************************************************************
***************************************************************************************************/
#include "mat.h"

/////////////////////////// VECTOR CLASS METHODS ///////////////////////////////

/**
****************************************************************************************************
@brief Construct vector from two points
@param start reference to first TVector point
@param end reference to second TVector point
***************************************************************************************************/
TVector::TVector(const TVector &start, const TVector &end)
{
  x = end.x - start.x;
  y = end.y - start.y;
  z = end.z - start.z;
  w = end.w - start.w;
}

/**
****************************************************************************************************
@brief Create vector from two points
@param start reference to first TVector point
@param end reference to second TVector point
***************************************************************************************************/
void TVector::Create(const TVector &start, const TVector &end)
{
  x = end.x - start.x;
  y = end.y - start.y;
  z = end.z - start.z;
  w = end.w - start.w;
}

/**
****************************************************************************************************
@brief Returns vector length
@return vector lenght as float number
***************************************************************************************************/
float TVector::Lenght()
{
  return (float)(sqrt(x*x + y*y + z*z + w*w));
}

/**
****************************************************************************************************
@brief Normalize vector values (sets values from 0.0 to 1.0)
***************************************************************************************************/
void TVector::Normalize()
{
  float lenght = Lenght();
  if(lenght==0) lenght=1;
  x /= lenght;
  y /= lenght;
  z /= lenght;
  w /= lenght;
}

/**
****************************************************************************************************
@brief Calculate dot product of two vectors
@param v1 reference to first TVector
@param v2 reference to second TVector
@return dot product as float number
***************************************************************************************************/
float DotProduct(const TVector &v1, const TVector &v2)
{
    return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z;
}

/**
****************************************************************************************************
@brief Calculate cross product of two vectors
@param v1 reference to first TVector
@param v2 reference to second TVector
@return cross product as new TVector
***************************************************************************************************/
TVector CrossProduct(const TVector &v1, const TVector &v2)
{
  TVector v;
  v.x = (v1.y * v2.z) - (v1.z * v2.y);
  v.y = (v1.z * v2.x) - (v1.x * v2.z);
  v.z = (v1.x * v2.y) - (v1.y * v2.x);
  return v;
}

/**
****************************************************************************************************
@brief Calculate angle of two vectors
@param v1 reference to first TVector
@param v2 reference to second TVector
@return vector angle in radians
***************************************************************************************************/
float VecAngle(TVector &v1, TVector &v2)
{
   float angle_cos = DotProduct(v1,v2)/( v1.Lenght() * v2.Lenght() );
   return acos(angle_cos);
   ///@todo make possible to return angle also in degrees
   //return acos(angle_cos)*180/PI;
}


/////////////////////////// MATRIX4X4 CLASS METHODS ////////////////////////////

/**
****************************************************************************************************
@brief Basic constructor, initialize all values to 0.0
***************************************************************************************************/
TMatrix::TMatrix()
{
	for(int i=0; i<16; i++) values[i] = 0.0;
}

/**
****************************************************************************************************
@brief Construct matrix directly from 16 float values
@param v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14,v15 float values
***************************************************************************************************/
TMatrix::TMatrix(   float v0,  float v1,  float v2,  float v3,
                    float v4,  float v5,  float v6,  float v7,
                    float v8,  float v9,  float v10, float v11,
                    float v12, float v13, float v14, float v15)
{
  values[0]=v0;   values[1]=v1;   values[2]=v2;   values[3]=v3;
  values[4]=v4;   values[5]=v5;   values[6]=v6;   values[7]=v7;
  values[8]=v8;   values[9]=v9;   values[10]=v10; values[11]=v11;
  values[12]=v12; values[13]=v13; values[14]=v14; values[15]=v15;
}

/**
****************************************************************************************************
@brief Copy constructor, initialize values from other TMatrix
@param ref reference matrix to copy from
***************************************************************************************************/
TMatrix::TMatrix(const TMatrix &ref)
{
  memcpy(values, ref.values, 16*sizeof(float));
}

/**
****************************************************************************************************
@brief Get one row from matrix
@param[in] position row position in the matrix
@param[out] ref pointer to float array, which will be filled with row values
***************************************************************************************************/
void TMatrix::GetRow(int position, float *ref) const
{
  ref[0] = values[position];
  ref[1] = values[4 + position];
  ref[2] = values[8 + position];
  ref[3] = values[12 + position];
}


/**
****************************************************************************************************
@brief create identity matrix
***************************************************************************************************/
void TMatrix::LoadIdentity()
{
  values[0]= 1.0; values[1]= 0.0; values[2]= 0.0; values[3]= 0.0;
  values[4]= 0.0; values[5]= 1.0; values[6]= 0.0; values[7]= 0.0;
  values[8]= 0.0; values[9]= 0.0; values[10]=1.0; values[11]=0.0;
  values[12]=0.0; values[13]=0.0; values[14]=0.0; values[15]=1.0;
}


/**
****************************************************************************************************
@brief Multiply current matrix with other through operator *
@param ref left-operator matrix
@return multiply result matrix
***************************************************************************************************/
TMatrix TMatrix::operator*(const TMatrix &ref) const
{
 TMatrix result;
 for(int i = 0; i < 4; i++)
   for(int j = 0; j < 4;j++)
   {
    result.values[4*i+j] = 0;
    for (int k = 0; k < 4;k++)
      result.values[4*i+j] += ref.values[4*i+k]*values[4*k+j];
   }
  return result;
}

/**
****************************************************************************************************
@brief Print out matrix values as std::ostream object
@param os reference to std::ostream
@param m reference to printed matrix
@return reference to new std::ostream object
***************************************************************************************************/
ostream& operator<<(ostream &os, const TMatrix &m)
{
  os  <<"| "<<m.values[0]<<", "<<m.values[1]<<", "<<m.values[2]<<", "<<m.values[3]<<"| "<<endl
      <<"| "<<m.values[4]<<", "<<m.values[5]<<", "<<m.values[6]<<", "<<m.values[7]<<"| "<<endl
      <<"| "<<m.values[8]<<", "<<m.values[9]<<", "<<m.values[10]<<", "<<m.values[11]<<"| "<<endl
      <<"| "<<m.values[12]<<", "<<m.values[13]<<", "<<m.values[14]<<", "<<m.values[15]<<"| "<<endl;
  return os;
}

/**
****************************************************************************************************
@brief returns inverse matrix from current 
@return inverse matrix
***************************************************************************************************/
TMatrix TMatrix::Inverse()
{
  TMatrix m(*this);
  
  // the inverse of the 3x3 rotation part is just the transpose
  for(int i=0;i<3;i++)
		for(int j=0;j<i;j++)
			std::swap(m.values[4 * i + j], m.values[4 * j + i]);

  // the inverse of the translation part is just the negative
  for(int i=0;i<3;i++)
  {
		m.values[4 * i + 3] = -m.values[4 * i + 3];
		m.values[12 + i] = -m.values[12 + i];
  }
  
  // and clear the translation part so that view is just a rotation matrix
  for(int i=0;i<3;i++) 
	{
		m.values[4 * i + 3] = 0;
		m.values[12 + i] = 0;  
  }
  return m;
}

/**
****************************************************************************************************
@brief returns transposed matrix from current
@return transposed matrix
***************************************************************************************************/
TMatrix TMatrix::Transpose()
{
    TMatrix tmp;
/*
    tmp[0][0] = values[0][0], tmp[0][1] = values[1][0], tmp[0][2] = values[2][0], tmp[0][3] = values[3][0];
    tmp[1][0] = values[0][1], tmp[1][1] = values[1][1], tmp[1][2] = values[2][1], tmp[1][3] = values[3][1];
    tmp[2][0] = values[0][2], tmp[2][1] = values[1][2], tmp[2][2] = values[2][2], tmp[2][3] = values[3][2];
    tmp[3][0] = values[0][3], tmp[3][1] = values[1][3], tmp[3][2] = values[2][3], tmp[3][3] = values[3][3];
*/
    tmp[0][0] = values[0], tmp[0][1] = values[4], tmp[0][2] = values[8], tmp[0][3] = values[12];
    tmp[1][0] = values[1], tmp[1][1] = values[5], tmp[1][2] = values[9], tmp[1][3] = values[13];
    tmp[2][0] = values[2], tmp[2][1] = values[6], tmp[2][2] = values[10], tmp[2][3] = values[14];
    tmp[3][0] = values[3], tmp[3][1] = values[7], tmp[3][2] = values[11], tmp[3][3] = values[15];
    return tmp;
}

