/**
****************************************************************************************************
****************************************************************************************************
@file mat.h
@brief Simple mathematical library to work with vectors and matrices
****************************************************************************************************
***************************************************************************************************/
#ifndef _MAT_H_
#define _MAT_H_
#include "globals.h"
///PI number, what to say more :)
#define PI 3.1415926535898f

///@class TVector 
///@brief basic vector class (vertex, normal...)
class TVector
{
  public:
    ///has four vector coordinates (can be XYZ position, RGB color...)
		///X - component
    GLfloat x;
		///Y - component
		GLfloat y;
		///Z - component
		GLfloat z;
		///W - component
		GLfloat w;        

    ///@brief Set null values
    TVector(GLfloat vx = 0.0,GLfloat vy = 0.0,GLfloat vz = 0.0, GLfloat vw = 0.0) 
        {x=vx; y=vy; z=vz; w=vw; }
        
    //makes vector from two points
    TVector(const TVector &start, const TVector &end);
    void Create(const TVector &start, const TVector &end);
    
    ///@brief Set values explicitly from floats
    void Set(GLfloat vx,GLfloat vy,GLfloat vz, GLfloat vw = 0.0)  
        {x=vx; y=vy; z=vz; w=vw; }
		///@brief null all vector values
    void LoadIdentity() 
        { x = 0.0; y = 0.0; z = 0.0; w = 0.0; }

    //returns vector length
    GLfloat Lenght();    
    //performs normalization of vector
    void Normalize();     
    ///@brief Return, if is vector empty
    bool IsEmpty() 
       { return (x == 0.0 && y == 0.0 && z == 0.0 && w == 0.0); }
    
    ///@brief retype operator
    operator GLfloat* () const {return (GLfloat*) this;}
		///@brief add constant to vector
		friend TVector operator + (const TVector &v, const float con)
		{ return TVector(v.x+con,v.y+con,v.z+con,v.w+con); }
		///@brief multiply by constant operator
		friend TVector operator * (const TVector &v, const float con)
		{ return TVector(v.x*con,v.y*con,v.z*con,v.w*con); }
		///@brief subtract two vectors
		friend TVector operator-(const TVector &v1, const TVector &v2)
		{ return TVector(v1.x-v2.x, v1.y-v2.y, v1.z-v2.z, v1.w-v2.z); }
		///@brief divide vector by constant
		friend TVector operator/(const TVector &v, const float con)
		{ return TVector(v.x/con, v.y/con, v.z/con, v.w/con); }
		///@brief invert values in vector
		friend TVector operator-(const TVector &v)
		{ return TVector(-v.x, -v.y, -v.z, -v.w); }
};



///@class TMatrix 
///@brief basic 4x4 Matrix class (row-major order OpenGL matrix)
class TMatrix
{
  private:
    ///has 16 float values
    float values[16];  
  public:
    //constructors
    TMatrix();
    TMatrix(  float v0,  float v1,  float v2,  float v3,
              float v4,  float v5,  float v6,  float v7,
              float v8,  float v9,  float v10, float v11,
              float v12, float v13, float v14, float v15);
    TMatrix(const TMatrix &ref);

    //get one row from matrix
    void GetRow(int position, float *ref) const;   
    //returns inverse matrix from current
    TMatrix Inverse();   
		TMatrix Transpose();
		//create identity matrix
		void LoadIdentity();
    
    //multiply of two matrices
    TMatrix operator*(const TMatrix &ref) const;   
    //prints matrix
    friend ostream& operator<<(ostream &os, const TMatrix &m); 

		///@brief reference operator
		///@return value at certain index
		float* operator [](int index)
		{ return &values[index]; }

    ///@brief retype operator, converts values to float array
    operator float* () const {return (float*) this;}
    ///@brief retype operator, converts values to constant float array
    operator const float* () const {return (const float*) this;}
};

////////////////////////////////////////////////////////////////////////////////

//cross product
TVector CrossProduct(const TVector &v1, const TVector &v2);
   
//dot product
float DotProduct(const TVector &v1, const TVector &v2);

//angle between two vectors
float VecAngle(TVector &v1, TVector &v2);

#endif
