/**
@file material.cpp
@brief material settings, dynamic shader creation
***************************************************************************************************/
#include "material.h"

/**
****************************************************************************************************
@brief Load custom shader from file
@param source shader source file 
@return string with shader source
***************************************************************************************************/
string LoadShader(const char* source)
{
  ifstream fin(source);
  if(!fin) 
  {
    string msg = "Cannot open shader ";
    msg += source;
    ShowMessage(msg.c_str(), false);
    return "null";
  }
  string data;
  char ch;
  while(fin.get(ch))
    data+=ch;

  return data;
}

////////////////////////////////////////////////////////////////////////////////
//************************* TMaterial methods ********************************//
////////////////////////////////////////////////////////////////////////////////


/**
****************************************************************************************************
@brief Direct creation of desired material
@param _name material name (must be unique)
@param amb ambient color (as RGB TVector)
@param diff diffuse color (as RGB TVector)
@param spec specular color (as RGB TVector)
@param shin shininess (0.0 - most shiny, 128.0 - least shiny)
@param reflect @todo material reflection
@param transp material transparency
@param lm light model (can be PHONG,GOURAUD,NONE)
***************************************************************************************************/
TMaterial::TMaterial(const char* _name, TVector amb, TVector diff, TVector spec, GLfloat shin, GLfloat reflect, GLfloat transp, GLint lm)
{
  name = _name;
  ambColor = amb;
  diffColor = diff;
  specColor = spec;
  shininess = shin;
  reflection = reflect;
  transparency = transp;
  lightModel = lm;

  shader = gshader_ref = 0;
  baked = false;
  useMRT = false;
  custom_shader = false;
  receive_shadows = true;
  use_gshader_ref = false;
}

/**
****************************************************************************************************
@brief Destroy material data, detache shaders
***************************************************************************************************/
TMaterial::~TMaterial()
{
	if(shader != 0)
	{
		glDetachObjectARB(shader,f_shader);
		glDetachObjectARB(shader,g_shader);
		glDetachObjectARB(shader,v_shader);
		glDeleteObjectARB(shader);
	}
}


/**
****************************************************************************************************
@brief Finds and returns next free texture name in list (important to synchronize generated texture names
to them stored in TMaterial::textures field)
@param texname suffix, texture type (like Base, Bump, Env...) 
@return new texture name
***************************************************************************************************/
string TMaterial::NextTexture(string texname)
{
  int i = 1;
  for(it = textures.begin(); it != textures.end(); it++)
  {
    ///find first empty texture, add index to it's name and return this name
    if(texname == it->first || it->first.empty() )
    {
      string rep;
			rep = num2str(char('A' + i));
      texname.replace(texname.length()-1, rep.size(), rep.c_str());
      i++;
    }
  }
  return texname;
}

/**
****************************************************************************************************
@brief Add texture from external file. If texture has been loaded, a texture pointer from cache is used
instead of reloading from file
@param file external texture file (.tga)
@param textype texture type (can be BASE,ENV,BUMP)
@param texmode texture addition mode (can be ADD,MODULATE,DECAL,BLEND,REPLACE)
@param intensity texture color intensity (or bump intensity)
@param tileX count of horizontal tiles
@param tileY count of vertical tiles
@param mipmap should we generate mipmaps for texture?
@param aniso should we use anisotropic filtering?
@param cache whether texture should be cached
@return new texture ID
****************************************************************************************************/
GLint TMaterial::AddTexture(const char *file, GLint textype, GLint texmode, GLfloat intensity, GLfloat tileX, GLfloat tileY, bool mipmap, bool aniso, GLint cache)
{
  ///1. generate new texture name (material name + texture mode (base, env, bump...) using TMaterial::NextTexture()
  string texname;
  switch(textype)
  {
    //base map
    case BASE: texname = NextTexture(name + "BaseA"); break;
    //alpha map
    case ALPHA: texname = NextTexture(name + "AlphaA"); break;
    //environment map
    case ENV: texname = NextTexture(name + "EnvA"); break;
    //bump map
    case BUMP: texname = NextTexture(name + "BumpA"); break;
    //parallax map
    case PARALLAX: texname = NextTexture(name + "ParallaxA"); break;   
    //displacement map
    case DISPLACE: texname = NextTexture(name + "DisplaceA"); break; 
    //cube map
    case CUBEMAP: texname = NextTexture(name + "CubeA"); break; 
    //environment cube map
    case CUBEMAP_ENV: texname = name + "CubeEnvA"; break;//NextTexture(name + "CubeEnvA"); break; 
    //render textures
    case RENDER_TEXTURE: texname = "renderTexture"; break; 
    case BLOOM_TEXTURE: texname = "bloomTexture"; break;
    case NORMAL_TEXTURE: texname = "normalTexture"; break;
    case BLUR_TEXTURE: texname = "blurTexture"; break;
   
    default: texname = NextTexture(name + "BaseA"); break;
  };

  ///2. load new texture into map array (using Texture::Load())
  return textures[texname].Load(texname.c_str(), textype, file, texmode, intensity, tileX, tileY, mipmap, aniso, cache);
}

/**
****************************************************************************************************
@brief Add cubemap texture from external file. If texture has been loaded, a texture pointer from cache is used
instead of reloading from file
@param files external texture files with every side of cube texture (.tga)
@param textype texture type (CUBEMAP)
@param texmode texture addition mode (can be ADD,MODULATE,DECAL,BLEND,REPLACE)
@param intensity texture color intensity (or bump intensity)
@param tileX count of horizontal tiles
@param tileY count of vertical tiles
@param aniso should we use anisotropic filtering?
@param cache whether texture should be cached
@return new texture ID
*****************************************************************************************************/
GLint TMaterial::AddTexture(const char **files, GLint textype, GLint texmode, GLfloat intensity, GLfloat tileX, GLfloat tileY, bool aniso, GLint cache)
{

  if(textype != CUBEMAP && textype != CUBEMAP_ENV)
  {
    cerr<<"ERROR: textype must be cubemap!\n";
    return -1;
  }

  ///1. generate new texture name (material name + texture mode (base, env, bump...) using TMaterial::NextTexture()
  string texname = NextTexture(name + "CubeA");

  ///2. load new texture into map array (using Texture::Load())
  return textures[texname].Load(texname.c_str(), textype, files, texmode, intensity, tileX, tileY, aniso, cache);
}



/**
****************************************************************************************************
@brief Add shadow map
@param map pointer to texture data
@param intensity shadow intensity (0 - transparent, 1 - opaque)
***************************************************************************************************/
void TMaterial::AddShadowMap(GLuint map, GLfloat intensity)
{
  //don't add shadow map to screen quad and for materials whose don't want to receive shadows!
  if(name == "screen_quad" || !receive_shadows) 
		return;
  
  ///1. generate new shadow texture name using TMaterial::NextTexture()
  string texname = NextTexture(name + "ShadowA");
  textures[texname].SetID(map);
  textures[texname].SetName(texname);
  textures[texname].SetType(SHADOW);

  ///@todo 2: if material is transparent, modify intensity of shadow
  if(transparency > 0.0) 
		intensity = transparency;

  textures[texname].SetIntensity(intensity);
}


/**
****************************************************************************************************
@brief Remove all shadow maps
***************************************************************************************************/
void TMaterial::RemoveShadows()
{
	vector<string> to_erase;

	for(it = textures.begin(); it != textures.end(); it++)
		if(strstr(it->first.c_str(),"Shadow") != NULL)
			to_erase.push_back(it->first);

	for(unsigned i=0; i<to_erase.size(); i++)
		textures.erase(to_erase[i]);

}


/**
****************************************************************************************************
@brief Render material. If hasn't been baked, bake him first(TMaterial::BakeMaterial())
@param render_all if we are rendering to shadow map, render material only if material is alpha-tested
***************************************************************************************************/
void TMaterial::RenderMaterial(bool render_all)
{
	//if we are rendering to shadow map, render material only if material is alpha-tested
	if(!render_all)	
	{
		bool found = false;
		for(it = textures.begin(); it != textures.end(); it++)
		{
			if(it->second.GetType() == ALPHA)
			{
				found = true;
				break;
			}
		}
		if(!found)
		{
			glUseProgram(0);
			return;
		}
	}
  ///enable shaders
  //normal or geometry shader?
  if(use_gshader_ref) 
		glUseProgram(gshader_ref);
  else 
		glUseProgram(shader);

  ///activate textures attached to material (Texture::ActivateTexture() )
  int i=0;
  for(it = textures.begin(); it != textures.end(); it++)
  {
    if(!it->second.Empty())
    {
			it->second.ActivateTexture(i);
      i++;
    }
  }
}

/**
****************************************************************************************************
@brief Set shadow map matrix for texture attached to material (Texture::SetShadowMatrix())
@param m shadow projection matrix
@param l_num light number (correction for multiple shadows)
***************************************************************************************************/
void TMaterial::SetShadowMatrix(TMatrix &m, GLint l_num)
{
  //don't add shadow map to screen quad
  if(name == "screen_quad") 
		return;
  
  string sh_str = name;
  sh_str += "Shadow";

  //obtain light number and update shadow matrix
	sh_str += num2str(char(l_num + 'A'));
  textures[sh_str].SetShadowMatrix(m);
}


/**
****************************************************************************************************
@brief Loads and sets custom shader from source file. Fragment and vertex shader is mandatory, 
geometry shader optional
@param vert_source source file for vertex shader
@param geom_source source file for geometry shader
@param frag_source source file for fragment shader
@param vert_defines additional strings for vertex shader (eg. "#define SOMETHING\n"), must be \n at the end
@param frag_defines additional strings for fragment shader (eg. "#define SOMETHING\n"), must be \n at the end
@param geom_defines additional strings for geometry shader (eg. "#define SOMETHING\n"), must be \n at the end
***************************************************************************************************/
bool TMaterial::CustomShader(const char* vert_source, const char* geom_source, const char* frag_source, 
                             const char *vert_defines, const char *frag_defines, const char *geom_defines)
{
  //create shaders
  v_shader = glCreateShader(GL_VERTEX_SHADER);
  f_shader = glCreateShader(GL_FRAGMENT_SHADER);

  string vertex_shader, fragment_shader;
  vertex_shader = "#version 150 compatibility\n";
  fragment_shader = "#version 150 compatibility\n";

  //additional defines
  if(vert_defines != NULL) 
		vertex_shader += vert_defines;
  if(frag_defines != NULL) 
		fragment_shader += frag_defines;

  //copy shader data
  vertex_shader += LoadShader(vert_source);
  fragment_shader += LoadShader(frag_source);

  if(vertex_shader == "null" || fragment_shader == "null")
	return false;

  const char *ff = fragment_shader.c_str();
  const char *vv = vertex_shader.c_str();

  //set shader source
  glShaderSource(v_shader, 1, &vv,NULL);
  glShaderSource(f_shader, 1, &ff,NULL);

  //compile and detect shader errors
  char log[BUFFER]; int len;
  glCompileShader(v_shader);
  glCompileShader(f_shader);

  //create and link shader program
  shader = glCreateProgram();
  glAttachShader(shader,f_shader);
  glAttachShader(shader,v_shader);

  //do we have geometry shader source file?
  if(geom_source != NULL)
  {
    //then load, create, compile and attach geometry shader
	g_shader = glCreateShader(GL_GEOMETRY_SHADER);
	string geometry_shader = "#version 150 compatibility\n";

	//additional defines
	if(geom_defines != NULL) 
	  geometry_shader += geom_defines;
	//load and link shader
    geometry_shader += LoadShader(geom_source);
	if(geometry_shader == "null")
		return false;
    const char *gg = geometry_shader.c_str();
    glShaderSource(g_shader, 1, &gg,NULL);
    glCompileShader(g_shader);
    glAttachShader(shader,g_shader);
  }

  //final shader linking
  glLinkProgram(shader);
  glUseProgram(shader);

  //shader creation status: print error if any
  glGetShaderInfoLog(v_shader, BUFFER, &len, log);
  if(strstr(log, "succes") == NULL && len > 0) 
		cout<<endl<<name<<":"<<log;    
  if(geom_source != NULL)
  {
    glGetShaderInfoLog(g_shader, BUFFER, &len, log);
    if(strstr(log, "succes") == NULL && len > 0) 
		cout<<endl<<name<<":"<<log;    
  }
  glGetShaderInfoLog(f_shader, BUFFER, &len, log);
  if(strstr(log, "succes") == NULL && len > 0) 
		cout<<endl<<name<<":"<<log;    


  //*************************************
  ///4 Get uniform variables for textures (using Texture::GetUniforms() )
	int i=0;
  for(it = textures.begin(); it != textures.end(); it++)
  {
    if(!it->second.Empty())
    {
      it->second.GetUniforms(shader);
	  it->second.ActivateTexture(i,true);
		i++;
    }
  }
#ifdef GL3
	//Bind generic vertex attributes: vertices, normals and coordinates
	glBindAttribLocation(shader, 0, "in_Vertex");
  glBindAttribLocation(shader, 1, "in_Normal");
	glBindAttribLocation(shader, 2, "in_Coord");
#endif

  gshader_ref = shader;   //duplicate to use for geometry shader

  ///set materials parameters
	SetUniforms("material.ambient",ambColor, 3);
	SetUniforms("material.diffuse",diffColor, 3);
	SetUniforms("material.specular",specColor, 3);
	SetUniform("material.shininess",shininess);

  baked = true;
  custom_shader = true;
	return true;
}


/**
****************************************************************************************************
@brief Set float uniform value in shader
@param v_name variable name
@param value variable value
***************************************************************************************************/
void TMaterial::SetUniform(const char* v_name, float value)
{
	if(!glIsProgram(shader)) 
		return;

	glUseProgram(shader);
	GLint loc = glGetUniformLocation(shader,v_name);
	if(loc < 0)
		return;
	glUniform1f(loc, value);
}

/**
****************************************************************************************************
@brief Set float/double uniform value in shader
@param v_name variable name
@param value variable value
***************************************************************************************************/
void TMaterial::SetUniform(const char* v_name, double value)
{
	if(!glIsProgram(shader)) 
		return;

	glUseProgram(shader);
	GLint loc = glGetUniformLocation(shader,v_name);
	if(loc < 0)
		return;
	glUniform1f(loc, (float)value);
}

/**
****************************************************************************************************
@brief Set int uniform value in shader
@param v_name variable name
@param value variable value
***************************************************************************************************/
void TMaterial::SetUniform(const char* v_name, int value)
{
	if(!glIsProgram(shader)) 
		return;

	glUseProgram(shader);
	GLint loc = glGetUniformLocation(shader,v_name);
	if(loc < 0)
		return;
	glUniform1i(loc, value);
}

/**
****************************************************************************************************
@brief Set multiple float uniform values in shader (vectors, matrices...)
@param v_name variable name
@param value variable values
@param values number of values
***************************************************************************************************/
void TMaterial::SetUniforms(const char* v_name, float *value, int values)
{
	if(!glIsProgram(shader)) 
		return;

	glUseProgram(shader);
	GLint loc = glGetUniformLocation(shader,v_name);
	if(loc < 0)
		return;

  //set vec2,vec3,vec4 or matrix variables
  switch(values)
  {
    case 2: glUniform2fv(loc, 1, value); break;
    case 3: glUniform3fv(loc, 1, value); break;
    case 4: glUniform4fv(loc, 1, value); break;
    case 9: glUniformMatrix3fv(loc, 1, 0, value); break;
    case 16: glUniformMatrix4fv(loc, 1, 0, value); break;
    default: break;
  }
}
