/**
****************************************************************************************************
****************************************************************************************************
@file: object.cpp
@brief objects classes, initialization and settings
****************************************************************************************************
***************************************************************************************************/
#include "object.h"

				////////////////////////////////////////////////////////////////////////////////
				//////////////////////////// TObject methods ///////////////////////////////////
				////////////////////////////////////////////////////////////////////////////////

/**
****************************************************************************************************
@brief Set default values, primitive type and resets position
****************************************************************************************************/
TObject::TObject()
{
  type = PRIMITIVE;
	//reset transformation
  pos.LoadIdentity();
  rot.LoadIdentity();
  scale.Set(1.0,1.0,1.0);
	transform.LoadIdentity();		//initialize transformation matrix
	transformed = false;

  mat = NULL;
  shadow_cast = true;
  draw_object = true;
	cf_buffer = cr_buffer = render_cubemap = render_cubemap_depth = 0;
	//drawmode
	instances = 1;
	drawmode = GL_TRIANGLES;
  vbo.indices = 0;
	vbo.vao = 0;

	sceneID = 0;
}


/**
****************************************************************************************************
@brief Destructor
@todo free dynamic data
****************************************************************************************************/
TObject::~TObject()
{
	//delete framebuffer, renderbuffer and cubemap (if object was reflective)
	if(HasFramebuffer())
	{
		glDeleteFramebuffers(1, &cf_buffer);
		glDeleteRenderbuffers(1, &cr_buffer);
		glDeleteTextures(1, &render_cubemap);
    glDeleteTextures(1, &render_cubemap_depth);
	}

  ///delete VBO buffers (only if object is not used in future)
	if(vbo.vao != 0)
	{
		glDeleteVertexArrays(1, &vbo.vao);
		glDeleteBuffers(4, vbo.buffer);
	}

}



/**
****************************************************************************************************
@brief Creates object using direct parameters. Objects are generated directly and then stored in VBO
for faster further drawing.
@param _name object name(must be unique)
@param primitive object type (can be CUBE,PLANE,CONE,CYLINDER,DISK,SPHERE,TORUS,TEAPOT,SPLINE,FONT)
@param size basic object size (depends on primitive type)
@param height object height (depends on primitive type)
@param sliceX polygon count horizontal (depends on primitive type)
@param sliceY polygon count vertical (depends on primitive type)
****************************************************************************************************/
void TObject::Create(const char* _name, int primitive, GLfloat size, GLfloat height, GLint sliceX, GLint sliceY)
{
  name = _name;
  scale.Set(1.0,1.0,1.0);
  mat = NULL;
  shadow_cast = true;
  refl_size = 256;
  draw_object = true;
	element_indices = true;
	transformed = false;

  //object type
  if(primitive == MY_POINT)
  {
    type = MY_POINT;
    return;
  }
  else
    type = PRIMITIVE;

  GLfloat x = size/2;
  GLfloat z = height/2;

  //prepare data buffers for object
  vector<float> vertices, normals,texcoords;
  vector<unsigned int> faces;

  ///Every primitive type use parameters differently:
  unsigned int verts = 0; //number of vertices
  switch(primitive)
  {
    ///cube: 1 param: size
    case CUBE:
      {
        verts = 20;
        vbo.indices = 36;
        drawmode = GL_TRIANGLES;

        float vertices_f[] = { -size,-size,-size, -size,-size,-size, -size,-size,-size, //0
															 -size,-size,size, -size,-size,size, -size,-size,size,		//3
															 size,-size,-size, size,-size,-size, size,-size,-size,		//6
															 size,-size,size, size,-size,size, size,-size,size,				//9
															 size,size,-size, size,size,-size,			//12
															 size,size,size, size,size,size,				//14
															 -size,size,-size, -size,size,-size,		//16
															 -size,size,size, -size,size,size				//18
                               };

        float texcoords_f[] = { 0.0,1.0, 1.0,0.0, 0.0,0.0, 
																0.0,0.0, 1.0,1.0, 0.0,1.0, 
																1.0,1.0, 0.0,0.0, 1.0,0.0,
																1.0,0.0, 0.0,1.0, 1.0,1.0,
																1.0,0.0, 0.0,0.0,
																1.0,1.0, 0.0,1.0,
																0.0,0.0, 1.0,0.0,
																0.0,1.0, 1.0,1.0 };

        float normals_f[] = { -1.0,-1.0,-1.0, -1.0,-1.0,-1.0, -1.0,-1.0,-1.0,
															 1.0,-1.0,1.0, 1.0,-1.0,1.0, 1.0,-1.0,1.0,
															 1.0,-1.0,-1.0, 1.0,-1.0,-1.0, 1.0,-1.0,-1.0,
															 1.0,-1.0,1.0, 1.0,-1.0,1.0, 1.0,-1.0,1.0,
															 1.0,1.0,-1.0, 1.0,1.0,-1.0,
															 1.0,1.0,1.0, 1.0,1.0,1.0,
															 -1.0,1.0,-1.0, -1.0,1.0,-1.0,
															 -1.0,1.0,1.0, -1.0,1.0,1.0 };

				unsigned int faces_f[] = { 11,5,2,  8,11,2,  14,10,7, 12,14,7, 19,15,13, 17,19,13, 4,18,16, 1,4,16,14,18,3,  3,9,14,  6,0,16,  12,6,16 };

        vertices = vector<float>(vertices_f, vertices_f + verts*3);
        texcoords = vector<float>(texcoords_f, texcoords_f + verts*2);
        normals = vector<float>(normals_f, normals_f + verts*3);
        faces = vector<unsigned int>(faces_f, faces_f + vbo.indices);
      }
      break;
    ///plane: 2 params, size, sizeY
    case PLANE:
      {
        verts = 4;
        vbo.indices = 4;
        drawmode = GL_TRIANGLE_STRIP;

        float vertices_f[] = { -x,0.0,z,   x,0.0,z,  -x,0.0,-z,  x,0.0,-z };
        float texcoords_f[] = { 0.0,0.0,  1.0,0.0,   0.0,1.0,  1.0,1.0 };
        float normals_f[] = { 0.0,1.0,0.0,  0.0,1.0,0.0,  0.0,1.0,0.0,  0.0,1.0,0.0 };
        unsigned short faces_f[] = { 0,1,2,3 };

        vertices = vector<float>(vertices_f, vertices_f + 12);
        texcoords = vector<float>(texcoords_f, texcoords_f + 8);
        normals = vector<float>(normals_f, normals_f + 12);
        faces = vector<unsigned int>(faces_f, faces_f + 4);
      }
      break;
    ///strip plane: 4 params, size, sizeY, sliceX, sliceY
    case STRIP_PLANE:
      {
				float dx = (float)size/sliceX;
				float dy = (float)height/sliceY;
				float dtx = 1.0f/sliceX;
				float dty = 1.0f/sliceY;

				//create vertices, normals and texcoords
				for(int i=0; i<sliceY; i++)
				{
					for(int j=0; j<sliceX; j++)
					{
						vertices.push_back(j*dx - size/2.0f); vertices.push_back(0.0f); vertices.push_back(i*dy - height/2.0f);
						normals.push_back(0.0); normals.push_back(1.0); normals.push_back(0.0); 
						texcoords.push_back(j*dtx); texcoords.push_back(i*dty); 
					}
				}
				//create indices
				for(int i=0; i<sliceY - 2; i++)
				{
					for(int j=0; j<sliceX - 2; j++)
					{
						int shift = i*sliceX;
						faces.push_back(shift + j); faces.push_back(shift + j + sliceX); faces.push_back(shift + j + sliceX + 1);
						faces.push_back(shift + j); faces.push_back(shift + j + sliceX + 1); faces.push_back(shift + j + 1);
					}
				}
				verts = sliceX * sliceY;
				vbo.indices = faces.size();
				drawmode = GL_TRIANGLES;
      }
      break;
    ///cone: 4 params: baseRadius, height, sliceX, sliceY
    case CONE:
        return;
    ///cylinder: 4 params: baseRadius, height, sliceX, sliceY
    case CYLINDER:

        return;
    ///disk: 4 params: innerRadius, outerRadius, sliceX, sliceY
    case DISK:
        return;
    ///sphere: 3 params: radius, sliceX, sliceY
    case SPHERE:
      {
        int rings = sliceX;
        int segments = sliceY;

        // set vertex count and index count 
        verts = (rings + 1)*(segments + 1);
        vbo.indices = 2 * rings * (segments + 1);
        drawmode = GL_TRIANGLE_STRIP;

        /////////////////////////////////////////////////////////////////////////////////////////////////
        //Generate sphere
        float fDeltaRingAngle = ( PI / rings );
        float fDeltaSegAngle = ( 2.0f * PI / segments );

        unsigned int wVerticeIndex = 0 ;
        // Generate the group of rings for the sphere
        for( int ring = 0; ring < rings + 1 ; ring++ )
        {
          float r0 = sin( ring * fDeltaRingAngle );
          float y0 = cos( ring * fDeltaRingAngle );

          // Generate the group of segments for the current ring
          for( int seg = 0; seg < segments + 1 ; seg++ )
          {
            float x0 = r0 * sin( seg * fDeltaSegAngle );
            float z0 = r0 * cos( seg * fDeltaSegAngle );

            vertices.push_back(x0*size); vertices.push_back(y0*size); vertices.push_back(z0*size);
            normals.push_back(x0); normals.push_back(y0); normals.push_back(z0);
            texcoords.push_back((float)seg/segments); texcoords.push_back((float)ring/rings);

            // add two indices except for last ring 
            if ( ring != rings ) 
            {
              faces.push_back(wVerticeIndex);
              faces.push_back(wVerticeIndex + (unsigned int)(segments + 1));
              wVerticeIndex ++ ; 
            } 
          }
        }
      }
      break;
    ///torus: 4 params: innerRadius, outerRadius, nsides, rings
    case TORUS:
        {
          ///(torus creation - thanks to http://www.sulaco.co.za/)
          float theta, phi, theta1,
                cosTheta, sinTheta,
                cosTheta1, sinTheta1,
                ringDelta, sideDelta,
                cosPhi, sinPhi, dist;
          vbo.indices = 0;

          sideDelta = 2.0f * PI / sliceX;
          ringDelta = 2.0f * PI / sliceY;

          theta = 0.0;
          cosTheta = 1.0;
          sinTheta = 0.0;

          for(int i = sliceY - 1; i >= 0; i--)
          {
            theta1 = theta + ringDelta;
            cosTheta1 = cos(theta1);
            sinTheta1 = sin(theta1);
            phi = 0.0;
            for(int j = sliceX; j >= 0; j--)
            {
              phi = phi + sideDelta;
              cosPhi = cos(phi);
              sinPhi = sin(phi);
              dist = height + (size * cosPhi);

              texcoords.push_back(cosTheta1 * cosPhi); texcoords.push_back(-sinTheta1 * cosPhi);
              normals.push_back(cosTheta1 * cosPhi); normals.push_back(-sinTheta1 * cosPhi); normals.push_back(sinPhi); 
              vertices.push_back(cosTheta1 * dist); vertices.push_back(-sinTheta1 * dist); vertices.push_back(size * sinPhi); 
              faces.push_back(vbo.indices++);

              texcoords.push_back(cosTheta * cosPhi); texcoords.push_back(-sinTheta * cosPhi);
              normals.push_back(cosTheta * cosPhi); normals.push_back(-sinTheta * cosPhi); normals.push_back(sinPhi); 
              vertices.push_back(cosTheta * dist); vertices.push_back(-sinTheta * dist); vertices.push_back(size * sinPhi); 
              faces.push_back(vbo.indices++);
            }
            theta = theta1;
            cosTheta = cosTheta1;
            sinTheta = sinTheta1;
          }

          verts = vbo.indices;
          drawmode = GL_TRIANGLE_STRIP;
        }
        break;  
    case SCREEN_QUAD:
        return;
    default:
        break;
  }
  //create vertex buffer with data
  glGenVertexArrays(1, &vbo.vao);
  glBindVertexArray(vbo.vao);

  //Allocate and assign three VBO to our handle
  glGenBuffers(4, vbo.buffer);

  //Bind our first VBO as being the active buffer and storing vertex attributes (coordinates) and copy buffer data
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_VERTEX]);
  glBufferData(GL_ARRAY_BUFFER, 3 * verts * sizeof(float), &vertices[0], GL_DYNAMIC_DRAW);  
#ifdef GL3
  // vertices are on index 0 and contains three floats per vertex
  glVertexAttribPointer(GLuint(0), 3, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(0);
#else
  glVertexPointer(3, GL_FLOAT, 0, 0);
  glEnableClientState(GL_VERTEX_ARRAY);
#endif

  //store normals
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_NORMAL]);    
  glBufferData(GL_ARRAY_BUFFER, 3 * verts * sizeof(float), &normals[0], GL_STATIC_DRAW);
#ifdef GL3
  // normals are on index 1 and contains three floats per vertex
  glVertexAttribPointer(GLuint(1), 3, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(1);
#else
  glNormalPointer(GL_FLOAT, 0, 0);
  glEnableClientState(GL_NORMAL_ARRAY);
#endif

  //store texture coordinates
  glBindBuffer(GL_ARRAY_BUFFER, vbo.buffer[P_TEXCOORD]);    
  glBufferData(GL_ARRAY_BUFFER, 2 * verts * sizeof(float), &texcoords[0], GL_STATIC_DRAW);
#ifdef GL3
  //coordinates are on index 2 and contains two floats per vertex
  glVertexAttribPointer(GLuint(2), 2, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(2);
#else
  glTexCoordPointer(2, GL_FLOAT, 0, 0);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
#endif

  //store vertex array indices
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo.buffer[P_INDEX]);
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, vbo.indices * sizeof(unsigned int), &faces[0], GL_STATIC_DRAW);
}

/**
****************************************************************************************************
@brief create object as instance from existing object
****************************************************************************************************/
void TObject::CreateInstance(const TObject &ref)
{
	*this = ref;

	type = INSTANCE;
}


/**
****************************************************************************************************
@brief Draws object by calling it's display list. Before this object is transformed and rotated to
it's current position. Then attached material is rendered(TMaterial::RenderMaterial())
There are specific types of settings, when object is not rendered (e.g. when rendering into shadow map
is active and object don't cast shadows)
@param viewMatrix actual view matrix
@param render_mat whether object material should be drawn
****************************************************************************************************/
void TObject::Draw(TMatrix &viewMatrix,  bool render_mat)
{
  //don't draw object with draw_object flag set to false
  if(!draw_object)
    return;

  if( (!render_mat && !shadow_cast) || name == "screen_quad")        //don't draw to shadow map or screen quad
    return;

  //object transformation
  glLoadMatrixf(viewMatrix);
  glMultMatrixf(transform);

  //render material(if enabled)
  if(mat != NULL && type != INSTANCE) 
		mat->RenderMaterial(render_mat);

  //one point only
  if(type == MY_POINT)
  {
    glBegin(GL_POINTS);
      glVertex4f(0.0,0.0,0.0,1.0);
    glEnd();
  }
  else
  {
		if(type != INSTANCE)
			glBindVertexArray(vbo.vao);
		//different drawing mode: simple vertex array or array with element indices
		if(element_indices) 
		{
			//it's possible to use geometry instancing (if there is more than 1 object instance)
			if(instances > 1)
				glDrawElementsInstancedEXT(drawmode, vbo.indices, GL_UNSIGNED_INT , 0, instances);
			else
				glDrawElements(drawmode, vbo.indices, GL_UNSIGNED_INT , 0);
		}
		else 
		{
			if(instances > 1)
				glDrawArraysInstancedEXT(GL_TRIANGLES, 0, vbo.indices * 3, instances);
			else
				glDrawArrays(GL_TRIANGLES, 0, vbo.indices * 3);
		}
  }
}

/**
****************************************************************************************************
@brief Draws screen aligned quad
****************************************************************************************************/
void TObject::DrawScreenQuad()
{
  //render material()
  mat->RenderMaterial();

  //draw quad
  glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(0.0,1.0); glVertex2f(-1.0, 1.0);
    glTexCoord2f(1.0,1.0); glVertex2f( 1.0, 1.0);
    glTexCoord2f(0.0,0.0); glVertex2f(-1.0,-1.0);
    glTexCoord2f(1.0,0.0); glVertex2f( 1.0,-1.0);
  glEnd();
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////// OBJECT TRANSFORMATIONS /////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
****************************************************************************************************
@brief Move object from current position by position vector(relative)
@param wx X-coordinate of vector
@param wy Y-coordinate of vector
@param wz Z-coordinate of vector
****************************************************************************************************/
void TObject::Move(GLfloat wx, GLfloat wy, GLfloat wz)
{
  pos.x += wx;
  pos.y += wy;
  pos.z += wz;

	//update matrix
	glLoadMatrixf(transform);
	glTranslatef(wx,wy,wz);
	glGetFloatv(GL_MODELVIEW_MATRIX,transform);
	transformed = true;
}

/**
****************************************************************************************************
@brief Move object to new position(relative)
@param wx X-coordinate of vector
@param wy Y-coordinate of vector
@param wz Z-coordinate of vector
****************************************************************************************************/
void TObject::MoveAbs(GLfloat wx, GLfloat wy, GLfloat wz)
{
  pos.x = wx;
  pos.y = wy;
  pos.z = wz;

	//update matrix
	glLoadIdentity();
	glTranslatef(pos.x,pos.y,pos.z);
	glRotatef(rot.x,1.0,0.0,0.0);
	glRotatef(rot.y,0.0,1.0,0.0);
	glRotatef(rot.z,0.0,0.0,1.0);
	glScalef(scale.x,scale.y,scale.z);
	glGetFloatv(GL_MODELVIEW_MATRIX,transform);
	transformed = true;
}

/**
****************************************************************************************************
@brief Rotate object by adding angle to current rotation(relative)
@param angle rotation increment in degrees
@param axis rotational axis (can be A_X, A_Y, A_Z)
****************************************************************************************************/
void TObject::Rotate(GLfloat angle, GLint axis)
{
	glLoadMatrixf(transform);
  switch(axis)
  {
    case A_X: 
				rot.x += angle;
				glRotatef(angle,1.0,0.0,0.0);
        break;
    case A_Y: 
				rot.y += angle;
				glRotatef(angle,0.0,1.0,0.0);
        break;
    case A_Z: 
				rot.z += angle;
				glRotatef(angle,0.0,0.0,1.0);
        break;
    default:
        break;
  }
	//update matrix	
	glGetFloatv(GL_MODELVIEW_MATRIX,transform);
	transformed = true;
}

/**
****************************************************************************************************
@brief Rotate object by new angle(absolute)
@param angle new rotation angle in degrees
@param axis rotational axis (can be A_X, A_Y, A_Z)
****************************************************************************************************/
void TObject::RotateAbs(GLfloat angle, GLint axis)
{
	glLoadIdentity();
  switch(axis)
  {
    case A_X: 
				rot.x = angle;
        break;
    case A_Y: 
				rot.y = angle;
        break;
    case A_Z: 
				rot.z = angle;
        break;
    default:
        break;
  }
	//update matrix	
	glTranslatef(pos.x,pos.y,pos.z);
	glRotatef(rot.x,1.0,0.0,0.0);
	glRotatef(rot.y,0.0,1.0,0.0);
	glRotatef(rot.z,0.0,0.0,1.0);
	glScalef(scale.x,scale.y,scale.z);
	glGetFloatv(GL_MODELVIEW_MATRIX,transform);
	transformed = true;
}

/**
****************************************************************************************************
@brief Resize object
@param sx scale in X-axis (<1.0 - smaller, >1.0 - bigger)
@param sy scale in Y-axis (<1.0 - smaller, >1.0 - bigger)
@param sz scale in Z-axis (<1.0 - smaller, >1.0 - bigger)
****************************************************************************************************/
void TObject::Resize(GLfloat sx, GLfloat sy, GLfloat sz)
{
  scale.x = sx;
  scale.y = sy;
  scale.z = sz;

	//update matrix
	glLoadMatrixf(transform);
	glScalef(sx,sy,sz);
	glGetFloatv(GL_MODELVIEW_MATRIX,transform);
}
