/**
****************************************************************************************************
****************************************************************************************************
@file: object.h
@brief objects classes, initialization and settings
Uses lib3DS to load 3D objects/scenes: http://code.google.com/p/lib3ds/
****************************************************************************************************
***************************************************************************************************/
#ifndef _OBJECT_H_
#define _OBJECT_H_

#include "material.h"
#include "../extern/lib3ds/lib3ds.h"

///Object types
enum Obj_types{PRIMITIVE,EXTERN,INSTANCE};
///Basic primitive types
enum Primitives{CUBE,PLANE,STRIP_PLANE,CONE,CYLINDER,DISK,SPHERE,TORUS,SPLINE,FONT,SCREEN_QUAD,MY_POINT};

///vertex buffer indices
enum VBOindices{P_VERTEX,P_NORMAL,P_TEXCOORD,P_INDEX};
///data structure to store vertex/texture/faces data
typedef float TVertex[3];
///data structure to store texture data
typedef float TCoord[2];

///@brief Vertex buffer objects structure
struct VBO{
	///pointers to all buffers: vertex, normal, texcoordinate and faces
  GLuint buffer[4];
	///pointer vertex array
  GLuint vao;
	///number of indice
  unsigned int indices;
};


/**
@class TObject
@brief contains object information(unique name, type), 3D data(vertices, faces, normal and texture coordinates),
position, rotation and scale, link to material and some specific settings (like shadow casting)
***************************************************************************************************/
class TObject
{
    private:
        TVector pos,rot,scale;    //position, rotation and scale
        bool draw_object;   //should be object drawn?

        //shadows settings
        bool shadow_cast;
        bool shadow_receive;

        //material properties
        TMaterial *mat;

				//transformation matrix
				TMatrix transform;
				bool transformed;

        int type;
        string name;
        VBO vbo;         //vertex buffer object
        GLenum drawmode; //VBO drawing mode
				bool element_indices;			//do we  have element indices instead of vertex array?

				//object instances used for geometry instancing
				int instances;

				//scene ID - when drawing more scenes than 1
				int sceneID;

    public:
			  ///framebuffer for dynamic reflection
        GLuint cf_buffer;
				///renderbuffer for dynamic reflection
				GLuint cr_buffer;
				///cubemap to render into
				GLuint render_cubemap;
				///depth texture for cubemap
				GLuint render_cubemap_depth;
				///reflection render size
				GLuint refl_size;

        //basic constructor
        TObject();
        //free dynamic data
        ~TObject();

        //create object as primitive
        void Create(const char *name, int primitive, GLfloat size, GLfloat height, GLint sliceX, GLint sliceY);
        //create object from 3DS file
        VBO Create(const char *name, const char *file, bool load = true);
        //create object from 3DS mesh
        VBO Create(Lib3dsMesh *mesh);
				//create object as instance from existing object
				void CreateInstance(const TObject &ref);
        ///set new VBO data
				///@param data already loaded VBO data
        void SetVBOdata(VBO data)
					{ vbo = data; }
				///@brief Set number of current object GI instances
				void SetGInstances(int count)
					{ if(count > 1) instances = count; }
				///@brief Return vertex buffer ID
				GLint GetVertexBuffer()
					{ return vbo.buffer[0]; }


        //draw object (with or without materials)
        void Draw(TMatrix &viewMatrix,  bool render_mat = true);
        //draw screen aligned quad
        void DrawScreenQuad();
        ///@brief turn on/off object drawing
        void DrawObject(bool flag)
          { draw_object = flag; }


        ///@brief Set material to object (material referenced as TMaterial class)
        void SetMaterial(TMaterial *m)
          { mat = m;}
        ///@brief Has object reflections?
        float GetReflect()
          { if(mat != NULL) return mat->GetReflect(); else return 0.0; }

        ///@brief Return if object has assigned material
        bool HasMaterial()
          { return !(mat == NULL); }
        ///@brief Return material name
        string GetMatName()
          { if(mat != NULL) return mat->GetName(); else return "<empty>"; }

        //object transformation
        void Move(GLfloat wx, GLfloat wy, GLfloat wz);
        void MoveAbs(GLfloat wx, GLfloat wy, GLfloat wz);
				///@brief Move object to new absolute position
				///@param w new position coordinates
        void MoveAbs(TVector w)
          { pos = w; }
        //object rotation
        void Rotate(GLfloat angle, GLint axis);
        void RotateAbs(GLfloat angle, GLint axis);
        //object resize
        void Resize(GLfloat sx, GLfloat sy, GLfloat sz);

        ///@brief Enable/disable shadow casting by object
        void CastShadow(bool flag = true)
          { shadow_cast = flag; }

        ///@brief Return object name
        string GetName()
          { return name; }
        ///@brief Return object position (as TVector)
        TVector GetPosition()
          {return pos;}
        ///@brief has object bound a transparent material?
        bool IsTransparent()
          {  if(mat != NULL) return (mat->GetTrasparency() > 0.0); else return 0; }
				///@brief has object bound framebuffer for reflection?
				bool HasFramebuffer()
				{ return (cf_buffer != 0 && cr_buffer != 0 && render_cubemap != 0); }

				///Get scene ID
				int GetSceneID()
				{  return sceneID; }
				///Set scene ID
				void SetSceneID(int id)
				{  sceneID = id; }
};
#endif
