/**
****************************************************************************************************
****************************************************************************************************
****************************************************************************************************
@mainpage gluxMark 2
@version 2.1
@author Juraj Vanek, xvanek29@stud.fit.vutbr.cz
@date 11.5.2009
@note gluxMark 2 - a program to test the performance of graphics accelerator using OpenGL/OpenCL.
Tests are designed to measure performance of every part of programmable graphics pipelina as well as
whole pipeline using advanced algoritms of realistic rendering.
Uses Simple DirectMedia Layer (http://www.libsdl.org/) and OpenGL Extension Wrangler (http://glew.sourceforge.net/)

@file main.cpp
@brief Entry point for application. Creates main OpenGL window using SDL, global scene settings,
tests invocation, starting point for application
****************************************************************************************************
****************************************************************************************************
***************************************************************************************************/
#include "test_fill.h"
#include "test_frag.h"
#include "test_geom.h"
#include "test_vert.h"
#include "test_complex.h"
#include "test_compute.h"

using namespace std;

///Current test
int curr_test;
///Final test
int final_test;
///Test class for initialization and run
Test *test;
///Total score
float total_score = 0.0;

///Drawn cycles in one second
unsigned cycle = 0;       
///Actual framerate
unsigned fps;

///Scene class for initialization and run
TScene *s;
///Text message
string msg;
///wireframe rendering flag
bool wire = false;                    

///stream to save results
stringstream results;

///Should we print only system info?
bool info = false;
///VRAM memory info
int meminfo[4];
///Screen width
int resx;
///Screen height
int resy;
///Antialiasing level
int fsaa = 1;

///developer mode
bool devmode = false;

/**
****************************************************************************************************
@brief Secure score by calculating its hash
@param str string with score number
@return hash as integer
****************************************************************************************************/
int CalcHash(string str)
{
  int ret = 0;
	for(unsigned i=0; i<str.length(); i++)
		ret += str[i]*17 + str[i]*57 + str[i]*93;

	return ret;
}

/**
****************************************************************************************************
@brief Write results to file
****************************************************************************************************/
void WriteResults()
{
	//no score in devmode
	if(devmode)
		total_score = 0;
	cout<<"TOTAL SCORE: "<<total_score<<endl;
	//write total score and end document
	results<<"<total_score>\n"<<total_score<<"</total_score>\n<check>"<<CalcHash(num2str(total_score))<<"</check>\n</root>\n";
  ofstream fout("results.xml");
  if(!fout)
  {
    ShowMessage("Cannot write results to file! (read-only filesystem?)",false);
#ifdef _WIN_
    ///@todo SDL_Quit crashes on Linux
    SDL_Quit();
#endif
    exit(1);
  }
  fout<<results.str();
  fout.close();
}

/**
****************************************************************************************************
@brief Skip to the next test
@return success/fail of new test creation
****************************************************************************************************/
bool NextTest()
{
  delete test;
	test = NULL;
	delete s;

  //disable FSAA at supersampling (there's no point on doing that)
  int msaa;
  if(curr_test == COMPLEX_SSAMPLING)
    msaa = 1;
  else
    msaa = fsaa;

	//Initialize new scene
	s = new TScene();
  if(!s->PreInit(resx, resy, 0.1f, 1000.0f, 45.0f,msaa))
    return false;

  try
  {
    //Test order and duration
    if(curr_test <= TEXEL_FILLRATE_FLOAT32)
      test = new TestFill(10000,curr_test,s);
    else if(curr_test == FRAG_LIGHTS)
      test = new TestFrag(30000,curr_test,s);
    else if(curr_test == FRAG_NOISE || curr_test == FRAG_REFRACT)
      test = new TestFrag(20000,curr_test,s);
    else if(curr_test == FRAG_PROCEDURAL || curr_test == FRAG_PARALLAX)
      test = new TestFrag(15000,curr_test,s);
    else if(curr_test == GEOM_TESSELATE)
      test = new TestGeom(20000,curr_test,s);
    else if(curr_test == GEOM_CUBEMAP)
      test = new TestGeom(20000,curr_test,s);
    else if(curr_test == GEOM_PARTICLES || curr_test == GEOM_INSTANCING)
      test = new TestGeom(20000,curr_test,s);
    else if(curr_test == VERT_LIGHTS)
      test = new TestVert(30000,curr_test,s);
    else if(curr_test == VERT_WAVES)
      test = new TestVert(15000,curr_test,s);
    else if(curr_test == VERT_DISPLACE)
      test = new TestVert(20000,curr_test,s);
    else if(curr_test >= COMPLEX_HDR && curr_test <= COMPLEX_SSAMPLING)
      test = new TestComplex(30000,curr_test,s);
#ifdef USE_CL
    else if(curr_test >= COMPUTE_NBODY)
      test = new TestCompute(15000,curr_test,s);
#endif
  }
  catch(int err)
  {
    delete test;
    test = NULL;
    curr_test++;
    return false;
  }
	curr_test++;
	return true;
}


/**
****************************************************************************************************
@brief Initialize scene
@return success/fail of initialization
****************************************************************************************************/
bool InitScene()
{

#ifdef _WIN_
	//turn off VSYNC
	typedef BOOL (APIENTRY *PFNWGLSWAPINTERVALFARPROC)( int );
	PFNWGLSWAPINTERVALFARPROC wglSwapIntervalEXT = 0;
  const char *extensions = (const char*)glGetString( GL_EXTENSIONS );

  if( strstr( extensions, "WGL_EXT_swap_control" ) == 0 )
    cout<<"Warning, vsync may be enabled."<<endl;
  else
  {
    wglSwapIntervalEXT = (PFNWGLSWAPINTERVALFARPROC)wglGetProcAddress( "wglSwapIntervalEXT" );

    if( wglSwapIntervalEXT )
      wglSwapIntervalEXT(0);
  }
#endif

  //begin creation of XML document
  results<<"<?xml version=\"1.0\"?>\n<root>\n";

	//write system info
	SystemInfo();

  //initialize first test
  NextTest();
  return true;
}


/**
****************************************************************************************************
@brief Print system info into XML file
****************************************************************************************************/
void SysInfo()
{
	//Initialize scene
	s = new TScene();
  s->PreInit(resx, resy, 0.1f, 1000.0f, 45.0f,false,false);

  //begin creation of XML document
  results<<"<?xml version=\"1.0\"?>\n<root>\n";

  SystemInfo();
  //print out system info to file
  results<<"</root>\n";

  ofstream fout;
  fout.open("sysinfo.xml");
  fout<<results.str();
  fout.close();
}


/**
****************************************************************************************************
@brief Main drawing function, runs the test
****************************************************************************************************/
void Redraw()
{
  test->Run();
  cycle++;                          //increment draw counter

  //FPS info
	msg = num2str(fps) + "  FPS";
	s->DrawScreenText(msg.c_str(),50.0,1.0,1.0);

	if(devmode)
	{
		//memory info (ATI only)
		glGetIntegerv(GL_TEXTURE_FREE_MEMORY_ATI,meminfo);
		msg = "VRAM used: " + num2str(1024 - meminfo[0]/1024) + " MB";
		s->DrawScreenText(msg.c_str(),8.0,1.0,1.0);
	}
}


/**
****************************************************************************************************
@brief Handle keyboard input
@param event keyboard event
****************************************************************************************************/
void KeyInput(SDL_Event event)
{
  switch(event.key.keysym.sym)
  {
    case SDLK_SPACE:         //space - skip test (devmode)
			  if(devmode)
				{
					//end after last test
					if(curr_test - 1  == final_test)
					{
						delete test;
						delete s;
						//ShowMessage("All tests finished.");
						WriteResults();
#ifdef _WIN_
						///@todo SDL_Quit crashes on Linux
						SDL_Quit();
#endif
						exit(0);
					}
					NextTest();
				}
        break;

    case SDLK_w:         //W - wireframe (devmode)
			  if(devmode)
				{
					wire = !wire;
					if(wire)
						glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
					else
						glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
				}
        break;


		case SDLK_LALT:
    case SDLK_ESCAPE:        //ESCAPE - abort test, ALT, WINKEY - forbidden keys
        delete test;
				test = NULL;
				delete s;
        //write partial results
        WriteResults();
        //ShowMessage("Test aborted");
#ifdef _WIN_
        ///@todo SDL_Quit crashes on Linux
        SDL_Quit();
#endif
        exit(0);
        break;
    default:
        break;
  }
}


///mouse button status
int status;                    
/**
****************************************************************************************************
@brief Handle mouse clicks
@param event mouse event
****************************************************************************************************/
void MouseClick(SDL_Event event)
{
  if(event.button.state = SDL_RELEASED)
    status = 0;
  else
    status = event.button.button;
}

/**
****************************************************************************************************
@brief Handle mouse motion
@param event mouse event
****************************************************************************************************/
void MouseMotion(SDL_Event event)
{
	if(devmode)
	{
		int x = event.motion.xrel;
		int y = event.motion.yrel;

		if(status == SDL_BUTTON_LEFT)   //rotate
		{
			s->RotateCamera((float)y, A_X);
			s->RotateCamera((float)x, A_Y);
		}
		else if(status == /* wtf??? proc nefuguje??SDL_BUTTON_RIGHT */ 4)  //zoom
			s->MoveCamera(0.0f,0.0f, x/10.0f );
		else if(status == SDL_BUTTON_MIDDLE)  //position
			s->MoveCamera( x/10.0f , -y/10.0f, 0.0f);

		//print out camera position
		//s->PrintCamera();
	}
}

/**
****************************************************************************************************
@brief Print program usage if wrong parameters have been used. Then exits program
****************************************************************************************************/
void WrongParams()
{
  cout<<"Wrong parameters.\n"
    "Usage: bin/bench.exe [-w|-f resX resY][-aa value]\n"
    "Parameters:\n"
    "-w,-f: windowed/fullscreen mode\n"
    "resX, resY: screen resolution in pixels\n"
    "-aa: antialiasing strength (0,1 = off)\n"
    "-t: test suite (can be \"all\", \"fillrate\", \"fragment\", \"geometry\", \"vertex\", \"complex\", \"compute\", \"info\")\n";
#ifdef _WIN_
  ///@todo SDL_Quit crashes on Linux
  SDL_Quit();
#endif
  exit(1);
}


/**
****************************************************************************************************
@brief Measure time in separate thread - solves problem when drawing is slower than 100fps
@param unused null parameter
****************************************************************************************************/
int Timer(void *unused)
{
  int msec = 0;

	while(true)
	{
		//animate scene
		if(test != NULL)
			test->Animate();
		msec++;

		//increment time after one second
    if(msec == 100)
    {
      //save FPS
      fps = cycle;
			if(test != NULL)
				test->UpdateFPS(fps);
      cycle = 0;
			msec = 0;
    }
		SDL_Delay(10);
	}
	return 0;
}



/**
****************************************************************************************************
///////////////////////////////////////////////////////////////////////////////////////////////////
@brief Main entry point for application - initialize render, create test and performs drawing loop
@param argc argument count
@param argv argument vector
@return exit code of application
///////////////////////////////////////////////////////////////////////////////////////////////////
****************************************************************************************************/
int main(int argc, char **argv)
{
  //application settings
  string param, test_suite = "all";
  bool fullscreen = false;
  resx = 1024;
	resy = 768;

  //parse parameters
  for(int i=1; i<argc; i++)
  {
    param = argv[i];

    ////////////////////////////////////////////////
    //fullscreen/windowed mode
    if(param == "-w" || param == "-f")
    {
      if(param == "-w")
        fullscreen = false;
      else if(param == "-f")
        fullscreen = true;
      //display resolution
      if(i+2 < argc)
      {
        resx = atoi(argv[i+1]);
        resy = atoi(argv[i+2]);
        i += 2;
      }
      else
        WrongParams();
    }
    ///////////////////////////////////////////////
    //antialiasing
    else if(param == "-aa")
    {
      if(i+1 < argc)
      {
        fsaa = atoi(argv[i+1]);
        i++;
      }
      else
        WrongParams();
    }
    ////////////////////////////////////////////
    //test suite
    else if(param == "-t")
    {
      if(i+1 < argc)
      {
        test_suite = argv[i+1];
        i++;
      }
      else
        WrongParams();
    }
		///////////////////////////////////////////
		//developer mode
		else if(param == "-dev")
			devmode = true;
    ///////////////////////////////////////////
    //error
    else
      WrongParams();
  }

  ////////////////////////////////////////////////////////////
  //Test settings
  //all tests
	int test_num = atoi(test_suite.c_str());

  if(test_suite == "all")
  {
    curr_test = TEXEL_FILLRATE_SINGLE;
#ifdef USE_CL
    final_test = COMPUTE_NBODY;
#else
		final_test = COMPLEX_SSAMPLING;
#endif
  }
  //fillrate tests
  else if(test_suite == "fillrate")
  {
    curr_test = TEXEL_FILLRATE_SINGLE;
    final_test = TEXEL_FILLRATE_FLOAT32;
  }
  //fragment shader tests
  else if(test_suite == "fragment")
  {
    curr_test = FRAG_LIGHTS;
    final_test = FRAG_REFRACT;
  }
  //geometry shader tests
  else if(test_suite == "geometry")
  {
    curr_test = GEOM_TESSELATE;
    final_test = GEOM_INSTANCING;
  }
  //vertex shader tests
  else if(test_suite == "vertex")
  {
    curr_test = VERT_LIGHTS;
    final_test = VERT_DISPLACE;
  }
  //complex tests
  else if(test_suite == "complex")
  {
    curr_test = COMPLEX_HDR;
    final_test = COMPLEX_SSAMPLING;
  }
#ifdef USE_CL
	//OpenCL compute tests
  else if(test_suite == "compute")
  {
    curr_test = COMPUTE_NBODY;
    final_test = COMPUTE_NBODY;
  }
#endif
  else if(test_suite == "info")
    info = true;
	//single tests by test number
	else if(test_num >= 0 && test_num < 21)
  {
    curr_test = test_num;
    final_test = test_num;
  }
  else
    WrongParams();

  //initialize SDL video
  if(SDL_Init(SDL_INIT_VIDEO) < 0)
		return 1;

#ifdef GL3
  //request an opengl 3.2 context
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);
#endif

  //enable double buffering with a 24bit Z buffer.
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

  //enable multisampling
  SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
  SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, fsaa);

  SDL_WM_SetCaption("gluxMark 2.0", 0);

  //info - only initialize OpenGL for info
  if(info)
  {
    cout<<"Gathering system info...\n";
    fullscreen = false;
    resx = resy = 1;
  }

  //set video mode
  int mode = SDL_OPENGL;
  if(fullscreen)
    mode |= SDL_FULLSCREEN;
  if(SDL_SetVideoMode(resx, resy, 32, mode) == NULL)
  {
    ShowMessage("Display mode not supported\n");
#ifdef _WIN_
    ///@todo SDL_Quit crashes on Linux
    SDL_Quit();
#endif
    return 1;
  }
  //hide cursor
  SDL_ShowCursor(SDL_DISABLE);

	//init GLEW
	if(glewInit() != GLEW_OK)
	{
    ShowMessage("GLEW initialization failed!\n");
#ifdef _WIN_
    ///@todo SDL_Quit crashes on Linux
    SDL_Quit();
#endif
    return 1;
	}

  //check for necessary extensions
  if(!GLEW_ARB_multitexture || !GLEW_ARB_vertex_shader || !GLEW_ARB_fragment_shader || !GLEW_ARB_multisample || !GLEW_EXT_framebuffer_object )
  {
		 cout<<"Your video card doesn't meet minimum requirements. You won't be able to run any of the test\n";
     return -1;
  }

	//system info
	if(info)
	{
		SysInfo();
#ifdef _WIN_
    ///@todo SDL_Quit crashes on Linux
    SDL_Quit();
#endif
		return 0;
	}

  //init scene
  if(!InitScene())
  {
#ifdef _WIN_
    ///@todo SDL_Quit crashes on Linux
    SDL_Quit();
#endif
		exit(1);
  }

	//save test settings
	results<<"<settings>\n"
		"<resolution>"<<resx<<"x"<<resy<<"</resolution>\n"
		"<antialias>"<<fsaa<<"</antialias>\n"
		"<series>"<<test_suite<<"</series>\n"
		"</settings>\n";

	//measure time in separate thread
	SDL_Thread *time_thread;
  time_thread = SDL_CreateThread(Timer, NULL);

  if(time_thread == NULL )
	{
    ShowMessage("Cannot create timer thread!\n");
#ifdef _WIN_
    ///@todo SDL_Quit crashes on Linux
    SDL_Quit();
#endif
    return 1;
  }

  //Main loop
  SDL_Event event;
  while(true)
  {
    //handle events with SDL
    if(SDL_PollEvent(&event))
    {
      if(event.type == SDL_QUIT)
        break;
      //keyboard events
      else if(event.type == SDL_KEYDOWN)
        KeyInput(event);
      //mouse events
      else if(event.type == SDL_MOUSEMOTION)
      {
        MouseMotion(event);
        MouseClick(event);
      }
    }

    //TIMELINE: run all tests after set period of time
    if(test == NULL || test->HasEnded())
    {
      //end after all test runned
      if(curr_test - 1  == final_test)
      {
				//fetch score from last test
        if(test != NULL)
				  total_score += test->GetScore();
        delete test;
				test = NULL;
        WriteResults();
        //END
#ifdef _WIN_
        ///@todo SDL_Quit crashes on Linux
        SDL_Quit();
#endif
        exit(0);
      }
			//fetch score from previous test
      if(test != NULL)
			  total_score += test->GetScore();
       //if init failed, continue to next test
      if(!NextTest())
        continue;
    }
    //call drawing function
    Redraw();
    SDL_GL_SwapBuffers();
  }

  //deinitialize SDL
#ifdef _WIN_
  ///@todo SDL_Quit crashes on Linux
  SDL_Quit();
#endif
  return 0;
}
