/***************************************************************************************************/
//
//  Program: gluxMark2, DIP
//  Author: Juraj Vanek, xvanek29@stud.fit.vutbr.cz
//  File: test_compute.cpp
//  Desc: implementation of OpenCL compute test series - function overriders,
//				set kernel arguments and kernel execution
//  Code adapted from: ATI Stream SDK, http://developer.amd.com/gpu/ATIStreamSDK/Pages/default.aspx, N-Body simulation
//
//**************************************************************************************************/

#include "test_compute.h"
//#include "demo.h"
#ifdef USE_CL

				//////////////////////// TEST1 - N-Body simulation ////////////////////////////

/****************************************************************************************************
Initialize test - set bodies count and allocate memory for application buffers
***************************************************************************************************/
bool NBodySim::Init()
{
	//initialize test data
  dT = 0.01f;
  espSqr = 300.0;
  initPos = NULL;
  initVel = NULL;
  vel = NULL;
	total_flops = 0;

	start = end = 0;
	kernel_time = flops = total_flops = 0.0;

	numBodies = BODIES;

  // make sure body count is multiple of group size
	if(numBodies < groupSize)
		numBodies = groupSize;
  numBodies = (numBodies / groupSize) * groupSize;

	//allocate host memory
  initPos = new cl_float[4 * numBodies];
  initVel = new cl_float[4 * numBodies];
	//allocate aligned block of memory for data
#if defined (_WIN32)
  pos = (cl_float*)_aligned_malloc(numBodies * sizeof(cl_float4), 16);
#else
  pos = (cl_float*)memalign(16, numBodies * sizeof(cl_float4));
#endif
  if(pos == NULL)
  { 
		ShowMessage("OpenCL: Failed to allocate host memory. (pos)", false);
    return false;
  }

#if defined (_WIN32)
  vel = (cl_float*)_aligned_malloc(numBodies * sizeof(cl_float4), 16);
#else
  vel = (cl_float*)memalign(16, numBodies * sizeof(cl_float4));
#endif
  if(vel == NULL)
  { 
    ShowMessage("OpenCL: Failed to allocate host memory. (vel)", false);
    return false;
  }

	// random initialization of inputs
  for(unsigned i = 0; i < numBodies; ++i)
  {
		//two different groups
		int shift, mass;
		if(i < numBodies/2)
		{
			mass = 500;
			shift = 0;
		}
		else
		{
			mass = 1000;
			shift = -40;
		}

    int index = 4 * i;
    // First 3 values are position in x,y and z direction
    for(int j = 0; j < 3; ++j)
      initPos[index + j] = float(rand()%50 + shift);
    // Mass value
    initPos[index + 3] = float(rand()%mass);
    // First 3 values are velocity in x,y and z direction
    for(int j = 0; j < 3; ++j)
      initVel[index + j] = 0.0f;
    // unused
    initVel[3] = 0.0f;
  }

  memcpy(pos, initPos, 4 * numBodies * sizeof(cl_float));
  memcpy(vel, initVel, 4 * numBodies * sizeof(cl_float));

	//Initialize OpenCL
	if(!InitCL())
		return false;

  //Create memory objects for position
  updatedPos = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, numBodies * sizeof(cl_float4), pos,  &status);
	assert(status == CL_SUCCESS);

  //Create memory objects for velocity
  updatedVel = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, numBodies * sizeof(cl_float4), vel, &status);
	assert(status == CL_SUCCESS);

  return true;
}

/****************************************************************************************************
Sets custom kernel arguments
***************************************************************************************************/
bool NBodySim::SetupKernelArgs()
{
	cl_int status;
  //Particle positions
  status = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*)&updatedPos);
	assert(status == CL_SUCCESS);

  //Particle velocity
  status = clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *)&updatedVel);
	assert(status == CL_SUCCESS);

  //numBodies
  status = clSetKernelArg(kernel, 2, sizeof(cl_int), (void *)&numBodies);
	assert(status == CL_SUCCESS);

  //Time step
  status = clSetKernelArg(kernel, 3, sizeof(cl_float), (void *)&dT);
	assert(status == CL_SUCCESS);

  //Upward Pseudoprobability
  status = clSetKernelArg(kernel, 4, sizeof(cl_float), (void *)&espSqr);
	assert(status == CL_SUCCESS);

  //Local memory
  status = clSetKernelArg(kernel,  5, groupSize * 4 * sizeof(float), NULL);
	assert(status == CL_SUCCESS);

  return true;
}

/****************************************************************************************************
Run kernel in whole N-D range
***************************************************************************************************/
bool NBodySim::Run()
{
  cl_event events[2];

  //Enqueue a kernel run call.
	size_t threads = (size_t)numBodies;
  status = clEnqueueNDRangeKernel(commandQueue, kernel, 1, NULL, &threads, &groupSize, 0, NULL, &events[0]);
	if(status != CL_SUCCESS)
	  return false;

  clFinish(commandQueue);
	//assert(status == CL_SUCCESS);

  //Wait for the kernel buffer to finish execution
  status = clWaitForEvents(1, &events[0]);

	clGetEventProfilingInfo(events[0], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
	clGetEventProfilingInfo(events[0], CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
	kernel_time = (end - start) * 1.0e-6f;
	flops = ( BODIES * BODIES * 21 * 1000.0f/kernel_time)/1000000000.0f;
	total_flops += flops;
  clReleaseEvent(events[0]);

  //Enqueue readBuffer
  status = clEnqueueReadBuffer(commandQueue, updatedPos, CL_TRUE, 0, numBodies* sizeof(cl_float4), pos, 0, NULL, &events[1]);

  //Wait for the read buffer to finish execution
  status = clWaitForEvents(1, &events[1]);
  clReleaseEvent(events[1]);

  return true;
}

/****************************************************************************************************
Free buffers and other program objects
***************************************************************************************************/
bool NBodySim::Destroy()
{
  cl_int status = TCompute::Destroy();
  status |= clReleaseMemObject(updatedPos);
  status |= clReleaseMemObject(updatedVel);

  //release program resources
  if(initPos)
    delete [] initPos;

  if(initVel)
    delete [] initVel;

  if(pos)
  {
#if defined (_WIN32)
    //_aligned_free(pos);
    free(pos);
#else
    free(pos);
#endif
  }
  if(vel)
  {
#if defined (_WIN32)
    //_aligned_free(vel);
    free(vel);
#else
    free(vel);
#endif
  }

	if(status == CL_SUCCESS)
		return true;
	else
		return false;
}

#endif
